
/*
* Copyright (c) 2008-2016 Allwinner Technology Co. Ltd.
* All rights reserved.
*
* File : vdecoder.c
* History :
*   Author  : xyliu <xyliu@allwinnertech.com>
*   Date    : 2016/04/13
*   Comment :
*
*
*/
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/unistd.h>
#include "vdecoder.h"
#include "sbmInterface.h"
#include "fbm.h"
#include "videoengine.h"
#include "ve_interface.h"
#include "mem_interface.h"
#include "log.h"
#include <linux/module.h>;
#include <linux/kernel.h>;

#define PIC_SAVE_FILE "/data/camera/picture.dat"
#define DEBUG_SAVE_BITSTREAM    (0)
#define DEBUG_SAVE_PICTURE      (1)

/* show decoder speed  */
#define AW_VDECODER_SPEED_INFO  (0)
#define DEBUG_MAX_FRAME_IN_LIST 16

#define CHECK_METADATA_INFO (0)
#define DECODER_SHOW_DETIAL_LOG (0)

//* setup the ve freq

#ifdef CONF_VE_FREQ_ENABLE_SETUP
#ifdef CONF_CERES_VE_FREQ_ENABLE_SETUP
#define VE_FREQ_SETTING_VALUE (576) //* just for h265 on a100
#else
#define VE_FREQ_SETTING_VALUE (648) //* just for h265 on h6
#endif
#endif

//* for check the proc info
#define ENABLE_SET_PROC_INFO (0)
#define SET_PROC_INFO_FREQ   (60)
#define PIC_ALIGN(value, n) (((value) + (n) - 1) & ~((n) - 1))

#define dy_log(fmt, arg...) do{ \
    if(p->bShowLogFlag == 1) \
        logd(fmt, arg); \
    else \
        logi(fmt, arg); \
}while(0)
typedef enum CdcSinkPicFormat {
    CDC_SINK_PIC_FORMAT_YU12,
    CDC_SINK_PIC_FORMAT_YV12,
    CDC_SINK_PIC_FORMAT_NV12,
    CDC_SINK_PIC_FORMAT_NV21,
    CDC_SINK_PIC_FORMAT_ARGB,
    CDC_SINK_PIC_FORMAT_RGB,
} CdcSinkPicFormat;

struct file *File;

typedef struct CdcSinkPicture {
    void* pData[4];
    int   nWidth;
    int   nHeight;
    int   nAlign;
    int   nBpp; //byte per pixle
    int   nAfbc;
    int   nAfbcSize;
    int   nTop;
    int   nBottom;
    int   nLeft;
    int   nRight;
    CdcSinkPicFormat nFormat;
} CdcSinkPicture;

typedef struct CdcBSHeader {
    char mMask[4];
    int  mBSIndex;
    int  mLen;
    u64  mPts;
} CdcBSHeader;

typedef struct CdcBitStream {
    void* pData;
    int   nLen;
    int   nAddHead;
    CdcBSHeader mHead;
} CdcBitStream;

typedef struct CdcSinkParam {
    //for all
    const char* pFileName;
    //for pic sink
    int   nLen;
    int   nStartNum;
    int   nToatalNum;
    int   bMd5Open;
} CdcSinkParam;
typedef struct CdcSinkInterface {
    //to picture
    int (*pWritePicture)(struct CdcSinkInterface *pSelf, CdcSinkPicture *pPic);
    int (*pWriteSinglePicture)(struct CdcSinkInterface *pSelf, CdcSinkPicture *pPic);
    //to bitstream 
    int (*pWriteBitStream)(struct CdcSinkInterface *pSelf, CdcBitStream* pStream);
    int (*pSetParameter)(struct CdcSinkInterface *pSelf, CdcSinkParam *pParam);
} CdcSinkInterface;

struct ScMemOpsS
{
    int  (*open)(void);
    int  (*open2)(void *, void *);
    void (*close)(void);

    int  (*total_size)(void);

    void *(*palloc)(int /*size*/, void *, void *);

    void *(*palloc_no_cache)(int /*size*/, void *, void *);

    void (*pfree)(void* /*mem*/, void *, void *);

    void (*flush_cache)(void * /*mem*/, int /*size*/);

    void *(*ve_get_phyaddr)(void * /*viraddr*/);

    void *(*ve_get_viraddr)(void * /*phyaddr*/);

    void *(*cpu_get_phyaddr)(void * /*viraddr*/);

    void *(*cpu_get_viraddr)(void * /*phyaddr*/);

    int (*mem_set)(void * /*s*/, int /*c*/, size_t /*n*/);

    int (*mem_cpy)(void * /*dest*/, void * /*src*/, size_t /*n*/);

    int (*mem_read)(void * /*dest */, void * /*src*/, size_t /*n*/);

    int (*mem_write)(void * /*dest*/, void * /*src*/, size_t /*n*/);

    int (*setup)(void);

    int (*shutdown)(void);

    unsigned int (*get_ve_addr_offset)(void);

    int (*get_debug_info)(char*, int);

    int (*get_vir_by_fd)(int/*dmafd*/, void* /*phy*/);
    int (*get_phy_by_fd)(int/*dmafd*/, void* /*phy*/);
    int (*free_phy_by_fd)(int/*dmafd*/, unsigned long /*phy*/);
    int (*get_fd_by_vir)(void * /*viraddr*/);
};

typedef struct SinkMD5
{
    u64 len;
    u8  block[64];
    u32 ABCD[4];
} SinkMD5;

typedef struct CdcPicSink {
    CdcSinkInterface mInterFace;
    char mFileName[30];
    struct file* pFile;
    int mStartNum;
    int mTotalNum;
    int mPicCount;
    int bMd5Open;
    SinkMD5 mMd5Ctx;
} CdcPicSink;
static inline int CdcSinkWritePic(struct CdcSinkInterface *pSelf, CdcSinkPicture *pPic)
{
        return pSelf->pWritePicture(pSelf, pPic);
}

static inline void CdcMemFlushCache(struct ScMemOpsS *memops, void* pMem, int nSize)
{
    //assert(memops != NULL);
    memops->flush_cache(pMem, nSize);
}

int file_write(struct file* file,unsigned long long offset, unsigned char* data, unsigned int size) {
    mm_segment_t oldfs;
    int ret;
    oldfs = get_fs();
    set_fs(KERNEL_DS);
    ret = vfs_write(file, data, size, &offset);
    set_fs(oldfs);
    
    return ret;

}
extern const char* strCodecFormat[];

const char* strDecodeResult[] =
{
    "UNSUPPORTED",
    "OK",
    "FRAME_DECODED",
    "CONTINU",
    "KEYFRAME_DECODED",
    "NO_FRAME_BUFFER",
    "NO_BITSTREAM",
    "RESOLUTION_CHANGE"
};

const char* strPixelFormat[] =
{
    "DEFAULT",          "YUV_PLANER_420",           "YUV_PLANER_420",
    "YUV_PLANER_444",   "YV12",                     "NV1",
    "YUV_MB32_420",     "MB32_422",                 "MB32_444",
};

typedef struct VideoDecoderDebugContext
{
    int      nFrameNum;
    int      nFrameNumDuration;
    int      nThreshHoldNum;
    float    nMaxCostTime;
    float    nCostTimeList[DEBUG_MAX_FRAME_IN_LIST];
    int64_t  nStartTime;
    int64_t  nDuration;
    int64_t  nCurrentTime;
    int64_t  nHardwareCurrentTime;
    int64_t  nHardwareDuration;
    int64_t  nHardwareTotalTime;
    int64_t  nGetOneFrameFinishTime;
    float    nExtraDecodeTimeList[DEBUG_MAX_FRAME_IN_LIST];
    float    nFrameTimeList[DEBUG_MAX_FRAME_IN_LIST];
    float    nFrameSmoothTime;
    int      nWaitingForDispPicNum[DEBUG_MAX_FRAME_IN_LIST];
    int      nValidDisplayPic;
    int      nStartIndexExtra;
    unsigned int *pFrameTime;
    int      nFrameTimeNum;
    unsigned int *pDispFrameNum;
    float    nMax_hardDecodeFps;
    float    nMin_hardDecodeFps;
}VideoDecoderDebugContext;

typedef struct VideoDecoderContext
{
    VConfig             vconfig;
    VideoStreamInfo     videoStreamInfo;
    VideoEngine*        pVideoEngine;
    int                 nFbmNum;
    Fbm*                pFbm[2];
    int                 nSbmNum;
    SbmInterface*       pSbm[2];
    VideoStreamDataInfo partialStreamDataInfo[2];

    VideoDecoderDebugContext debug;
    struct mem_interface *memops;
    struct ve_interface  *veops;
    int                 nIonFd;
    int                 nSavePicCount;
    int                 nDecFrameCount;

    int                 bSaveStreamFlag;
    int                 bSavePicFlag;
    int                 nSavePicStartNum;
    int                 nSavePicToatalNum;
    const char*         pSavePicPath;
    int                 bSaveSpecialStreamFlag;
    const char*         pSaveSpecialPath;

    int                 bShowSpeedInfo;
    int                 bShowLogFlag;
    CdcSinkInterface*   pPicSink;
}VideoDecoderContext;

static int  DecideStreamBufferSize(VideoDecoderContext* p);
static void UpdateVideoStreamInfo(VideoDecoderContext* p, VideoPicture* pPicture);
extern void ConvertPixelFormat(VideoPicture* pPictureIn, VideoPicture* pPictureOut);
extern int  RotatePicture0Degree(VideoPicture* pPictureIn,
                             VideoPicture* pPictureOut,
                             int nGpuYAlign,
                             int nGpuCAlign);
extern int  RotatePicture90Degree(VideoPicture* pPictureIn, VideoPicture* pPictureOut);
extern int  RotatePicture180Degree(VideoPicture* pPictureIn, VideoPicture* pPictureOut);
extern int  RotatePicture270Degree(VideoPicture* pPictureIn, VideoPicture* pPictureOut);
int CheckSbmDataFirstLastFlag(VideoDecoder*        pDecoder,
                          VideoStreamDataInfo* pDataInfo,
                          int                  nStreamBufIndex);

#if(CHECK_METADATA_INFO)
static void printfMetaDataInfo(struct sunxi_metadata* pSunxi_metadata)
{

    logd("metadata--hdr: d_x: %d, %d, %d, d_y: %d, %d, %d, w_x = %d, w_y = %d,\
maxDML = %d, minDML = %d",
         pSunxi_metadata->hdr_smetada.disp_master.display_primaries_x[0],
         pSunxi_metadata->hdr_smetada.disp_master.display_primaries_x[1],
         pSunxi_metadata->hdr_smetada.disp_master.display_primaries_x[2],
         pSunxi_metadata->hdr_smetada.disp_master.display_primaries_y[0],
         pSunxi_metadata->hdr_smetada.disp_master.display_primaries_y[1],
         pSunxi_metadata->hdr_smetada.disp_master.display_primaries_y[2],
         pSunxi_metadata->hdr_smetada.disp_master.white_point_x,
         pSunxi_metadata->hdr_smetada.disp_master.white_point_y,
         pSunxi_metadata->hdr_smetada.disp_master.max_display_mastering_luminance,
         pSunxi_metadata->hdr_smetada.disp_master.min_display_mastering_luminance);
    logd("metadata--hdr: maxCLL = %d, maxFALL = %d",
         pSunxi_metadata->hdr_smetada.maximum_content_light_level ,
         pSunxi_metadata->hdr_smetada.maximum_frame_average_light_level);

    char* pSignature = (char*)&pSunxi_metadata->afbc_head.signature;
    logd("metadata-afbc: signature = %c%c%c%c, filehdrSize = %d, version = %d, bodySize = %d",
        pSignature[0],pSignature[1],pSignature[2],pSignature[3],
        pSunxi_metadata->afbc_head.filehdr_size,
        pSunxi_metadata->afbc_head.version,
        pSunxi_metadata->afbc_head.body_size);
    logd("metadata-afbc: ncomponents = %d, header_layout = %d, yuvTran = %d, blockSplit = %d",
        pSunxi_metadata->afbc_head.ncomponents,
        pSunxi_metadata->afbc_head.header_layout,
        pSunxi_metadata->afbc_head.yuv_transform,
        pSunxi_metadata->afbc_head.block_split);
    logd("metadata-afbc: inputbits: %d, %d, %d, %d",
        pSunxi_metadata->afbc_head.inputbits[0],
        pSunxi_metadata->afbc_head.inputbits[1],
        pSunxi_metadata->afbc_head.inputbits[2],
        pSunxi_metadata->afbc_head.inputbits[3]);
    logd("metadata-afbc: block_w = %d, block_h = %d, w = %d, h = %d",
        pSunxi_metadata->afbc_head.block_width,
        pSunxi_metadata->afbc_head.block_height,
        pSunxi_metadata->afbc_head.width,
        pSunxi_metadata->afbc_head.height);
    logd("metadata-afbc: left_crop = %d, top_crop = %d, block_layout = %d",
        pSunxi_metadata->afbc_head.left_crop,
        pSunxi_metadata->afbc_head.top_crop,
        pSunxi_metadata->afbc_head.block_layout);

};
#endif
int PicSinkWritePic(CdcSinkInterface* pSelf, CdcSinkPicture* pPic)
{
    int nLen;

    CdcPicSink* pContext = (CdcPicSink*)pSelf;
    if(pPic == NULL || pContext->pFile == NULL)
    {
        return -1;
    }

    if(pContext->mPicCount < pContext->mStartNum ||
       pContext->mPicCount >= pContext->mStartNum + pContext->mTotalNum)
    {
        return 0;
    }

//    if(pContext->bMd5Open)
  //  {
    //    InterPicSaveMd5(pContext, pPic);
     //   goto EXIT;
  //  }
    if(pPic->nAfbc)
    {
        file_write(pPic->pData[0], 1, pPic->nAfbcSize, pContext->pFile);
        goto EXIT;
    }

    nLen = pPic->nWidth * pPic->nHeight;
    switch(pPic->nFormat)
    {
        case CDC_SINK_PIC_FORMAT_NV12:
        case CDC_SINK_PIC_FORMAT_NV21:
            if(pPic->nBpp == 10)
                nLen *= 2;
            file_write(File, 0, nLen, pPic->pData[0]);
            file_write(File, 0, nLen/2, pPic->pData[1]);
            break;
        default:
            loge("picture format not support");
    }

EXIT:
    logd("save pic width:%d height:%d format:%d", pPic->nWidth, pPic->nHeight, pPic->nFormat);
    pContext->mPicCount++;

    return 0;
}
void DebugSavePicture(VideoDecoderContext* p, VideoPicture* pPicture)
{
    CdcSinkPicture mPic;

    if(pPicture->pData0 == NULL)
    {
        logw("save picture failed: pPicture->pData0 is null");
        return ;
    }
    if(p->pPicSink != NULL)
    {
        switch(pPicture->ePixelFormat)
        {
            case PIXEL_FORMAT_YV12:
            case PIXEL_FORMAT_YUV_PLANER_420:
                if(pPicture->ePixelFormat == PIXEL_FORMAT_YUV_PLANER_420)
                    mPic.nFormat = CDC_SINK_PIC_FORMAT_YU12;
                else
                    mPic.nFormat = CDC_SINK_PIC_FORMAT_YV12;
                mPic.pData[0] = pPicture->pData0;
                mPic.pData[1] = (char*)mPic.pData[0] + pPicture->nWidth * pPicture->nHeight;
                mPic.pData[2] = (char*)mPic.pData[1] + pPicture->nWidth * pPicture->nHeight/2;
                break;
            case PIXEL_FORMAT_NV12:
            case PIXEL_FORMAT_NV21:
                if(pPicture->ePixelFormat == PIXEL_FORMAT_NV12)
                    mPic.nFormat = CDC_SINK_PIC_FORMAT_NV12;
                else
                    mPic.nFormat = CDC_SINK_PIC_FORMAT_NV21;
                mPic.pData[0] = pPicture->pData0;
                mPic.pData[1] = (char*)mPic.pData[0] + pPicture->nWidth * pPicture->nHeight;
                break;
            default:
                logw("not support now");
        }
        mPic.nWidth  = pPicture->nWidth;
        mPic.nHeight = pPicture->nHeight;
        mPic.nAlign  = 16;
        mPic.nBpp    = pPicture->b10BitPicFlag;
        mPic.nAfbc   = pPicture->bEnableAfbcFlag;
        mPic.nAfbcSize = pPicture->nAfbcSize;
        mPic.nTop    = pPicture->nTopOffset;
        mPic.nBottom = pPicture->nBottomOffset;
        mPic.nLeft   = pPicture->nLeftOffset;
        mPic.nRight  = pPicture->nRightOffset;
        CdcMemFlushCache((struct ScMemOpsS *)(p->memops), pPicture->pData0, mPic.nWidth*mPic.nHeight*3/2);
        PicSinkWritePic(p->pPicSink, &mPic);
    }
}

int PicSinkSetParam(CdcSinkInterface* pSelf, CdcSinkParam* pParam)
{
    CdcPicSink* pContext = (CdcPicSink*)pSelf;
    if(pParam == NULL)
    {
        loge("param is null");
        return -1;
    }

    memcpy(pContext->mFileName, pParam->pFileName, pParam->nLen);
    pContext->mStartNum = pParam->nStartNum;
    pContext->mTotalNum = pParam->nToatalNum;
    pContext->bMd5Open = pParam->bMd5Open;
    pContext->pFile = filp_open(pContext->mFileName, O_RDWR, 0);
    if(pContext->pFile == NULL)
    {
        loge("open file failure, name:%s", pContext->mFileName);
        return -1;
    }
    return 0;
}

VideoDecoder* CreateVideoDecoder(void)
{


    VideoDecoderContext* p;
    CdcSinkParam nSinkParm;
	struct mem_param param;
	struct ve_config mVeConfig;
    logv("CreateVideoDecoder ****");

    p = (VideoDecoderContext*)kmalloc(sizeof(VideoDecoderContext), GFP_KERNEL | __GFP_ZERO);
    if(p == NULL)
    {
        loge("memory alloc fail.");
        return NULL;
    }

    memset(p, 0, sizeof(VideoDecoderContext));

    //* the partialStreamDataInfo is for data submit status recording.
    p->partialStreamDataInfo[0].nStreamIndex = 0;
    p->partialStreamDataInfo[0].nPts         = -1;
    p->partialStreamDataInfo[1].nStreamIndex = 1;
    p->partialStreamDataInfo[1].nPts         = -1;
	
    p->memops = mem_create(MEM_TYPE_ION, param);
	if(p->memops == NULL)
    {
        loge("get ve ops failed , type: VE_TYPE_AW");
        kfree(p);
        return NULL;
    }
    mVeConfig.dec_flag = 1;
    mVeConfig.enc_flag = 0;
    mVeConfig.just_isp_enable = 0;
	mVeConfig.codec_format = VIDEO_CODEC_MJPEG; 
    p->veops = ve_create(VE_TYPE_AW, mVeConfig);
//	loge("ji*******ve create:%x", p->veops);
    if(p->veops == NULL)
    {
        loge("get ve ops failed , type: VE_TYPE_AW");
        kfree(p);
        return NULL;
    }
    int bSavePicFlag = 0;
#if DEBUG_SAVE_PICTURE
//    bSavePicFlag = 1;
#endif
        if(bSavePicFlag == 1)
    {
        nSinkParm.nStartNum = 0;
        nSinkParm.nToatalNum = 10;
        nSinkParm.pFileName = "/picture.dat";
        if(nSinkParm.pFileName == NULL)
        {
            nSinkParm.nStartNum = 0;
            nSinkParm.nToatalNum = 10;
            nSinkParm.pFileName = PIC_SAVE_FILE;
            nSinkParm.bMd5Open = 1;
        }
        
            nSinkParm.nLen = strlen(nSinkParm.pFileName) + 1;
            File == filp_open(nSinkParm.pFileName,O_RDWR,0);
            if(File == NULL)
    {
        loge("open file failure, name:%s", nSinkParm.pFileName);
        return -1;
    }
           // PicSinkSetParam(p->pPicSink, &nSinkParm);
        
    }
    return (VideoDecoder*)p;
}
EXPORT_SYMBOL(CreateVideoDecoder);

void DestroyVideoDecoder(VideoDecoder* pDecoder)
{
    VideoDecoderContext* p;

    logv("DestroyVideoDecoder, pDecoder=%p", pDecoder);

    p = (VideoDecoderContext*)pDecoder;

    //* Destroy the video engine first.
    if(p->pVideoEngine != NULL)
    {
        VideoEngineDestroy(p->pVideoEngine);
    }

    //* the memory space for codec specific data was allocated in InitializeVideoDecoder.
    if(p->videoStreamInfo.pCodecSpecificData != NULL)
        kfree(p->videoStreamInfo.pCodecSpecificData);

   	mem_destory(MEM_TYPE_ION, p->memops);

    ve_destory(VE_TYPE_AW, p->veops);
    kfree(p);
    return;
}
EXPORT_SYMBOL(DestroyVideoDecoder);

int InitializeVideoDecoder(VideoDecoder* pDecoder, VideoStreamInfo* pVideoInfo,  VConfig* pVconfig)
{
    VideoDecoderContext* p;
    int                  nStreamBufferSize;
    //int                  i;
    p = (VideoDecoderContext*)pDecoder;

    #ifdef CONF_VE_FREQ_ENABLE_SETUP
    if(pVconfig->nVeFreq == 0 && pVideoInfo->eCodecFormat == VIDEO_CODEC_FORMAT_H265)
        pVconfig->nVeFreq = VE_FREQ_SETTING_VALUE;
    #endif

    if(pVconfig->bThumbnailMode == 0)
    {
        logd("*** pVconfig->nVeFreq = %d",pVconfig->nVeFreq);
    }

    if(pVconfig->bSetProcInfoEnable == 0)
    {
        pVconfig->bSetProcInfoEnable = ENABLE_SET_PROC_INFO;
        pVconfig->nSetProcInfoFreq   = SET_PROC_INFO_FREQ;
    }
    else
    {
        if(pVconfig->nSetProcInfoFreq == 0)
            pVconfig->nSetProcInfoFreq   = SET_PROC_INFO_FREQ;
    }
    if(pVconfig->bSecureosEn == 1)
        pVconfig->bSetProcInfoEnable = 0;

    //* check codec format.
    if(pVideoInfo->eCodecFormat > VIDEO_CODEC_FORMAT_MAX ||
        pVideoInfo->eCodecFormat <  VIDEO_CODEC_FORMAT_MIN)
    {
        loge("codec format(0x%x) invalid.", pVideoInfo->eCodecFormat);
        return -1;
    }

    if(pVideoInfo->nHeight > 3840 || pVideoInfo->nWidth > 6144)//>6k
    {
        loge("pVideoInfo->nWidth = %d, pVideoInfo->nHeight = %d unsupported!!",
             pVideoInfo->nWidth, pVideoInfo->nHeight);
        return VDECODE_RESULT_UNSUPPORTED;
    }

    //* mjpeg decoder just support 16-align
    if(pVideoInfo->eCodecFormat == VIDEO_CODEC_FORMAT_MJPEG
        && pVconfig->nAlignStride > 16)
    {
        int nWidth = pVideoInfo->nWidth;
        if(pVconfig->bScaleDownEn == 1 && pVconfig->nHorizonScaleDownRatio !=0)
            nWidth = pVideoInfo->nWidth >> pVconfig->nHorizonScaleDownRatio;

        if(nWidth > 0 && (nWidth%pVconfig->nAlignStride) != 0)
        {
            logw("change nAlignStride from %d to 16, \
            as width of mjpeg decoder just support 16-align", pVconfig->nAlignStride);
            pVconfig->nAlignStride = 16;
        }
    }

    if(pVconfig->memops)
	{
		loge("you should not set memops");
		return -1;
	}
    
	
    //* print video stream information.
    {
        logv("Video Stream Information:");
        logv("     codec          = %s",
            strCodecFormat[pVideoInfo->eCodecFormat - VIDEO_CODEC_FORMAT_MIN]);
        logv("     width          = %d pixels", pVideoInfo->nWidth);
        logv("     height         = %d pixels", pVideoInfo->nHeight);
        logv("     frame rate     = %d",        pVideoInfo->nFrameRate);
        logv("     frame duration = %d us",     pVideoInfo->nFrameDuration);
        logv("     aspect ratio   = %d",        pVideoInfo->nAspectRatio);
        logv("     is 3D stream   = %s",        pVideoInfo->bIs3DStream ? "yes" : "no");
        logv("     csd data len   = %d",        pVideoInfo->nCodecSpecificDataLen);
        logv("     bIsFrameCtsTestFlag  = %d",  pVideoInfo->bIsFrameCtsTestFlag);
    }

    //* save the video stream information.
    p->videoStreamInfo.eCodecFormat          = pVideoInfo->eCodecFormat;
    p->videoStreamInfo.nWidth                = pVideoInfo->nWidth;
    p->videoStreamInfo.nHeight               = pVideoInfo->nHeight;
    p->videoStreamInfo.nFrameRate            = pVideoInfo->nFrameRate;
    p->videoStreamInfo.nFrameDuration        = pVideoInfo->nFrameDuration;
    p->videoStreamInfo.nAspectRatio          = pVideoInfo->nAspectRatio;
    p->videoStreamInfo.bIs3DStream           = pVideoInfo->bIs3DStream;
    p->videoStreamInfo.bIsFramePackage       = pVideoInfo->bIsFramePackage;
    p->videoStreamInfo.nCodecSpecificDataLen = pVideoInfo->nCodecSpecificDataLen;
    p->videoStreamInfo.bSecureStreamFlag     = pVideoInfo->bSecureStreamFlag;
    p->videoStreamInfo.bSecureStreamFlagLevel1 = pVideoInfo->bSecureStreamFlagLevel1;
    p->videoStreamInfo.bIsFrameCtsTestFlag = pVideoInfo->bIsFrameCtsTestFlag;

    if(p->videoStreamInfo.nCodecSpecificDataLen > 0)
    {
        int   nSize = p->videoStreamInfo.nCodecSpecificDataLen;
        char* pMem  = (char*)kmalloc(nSize, GFP_KERNEL | __GFP_ZERO);

        if(pMem == NULL)
        {
            p->videoStreamInfo.nCodecSpecificDataLen = 0;
            loge("memory alloc fail.");
            return -1;
        }

        memcpy(pMem, pVideoInfo->pCodecSpecificData, nSize);
        p->videoStreamInfo.pCodecSpecificData = pMem;
    }

    if(pVconfig->nDecodeSmoothFrameBufferNum <= 0
       || pVconfig->nDecodeSmoothFrameBufferNum > 16)
    {
        if(pVconfig->bThumbnailMode == 0)
        {
            logw("warning: the nDecodeSmoothFrameBufferNum is %d",
            pVconfig->nDecodeSmoothFrameBufferNum);
        }
    }
    if(pVconfig->nDeInterlaceHoldingFrameBufferNum <= 0
       || pVconfig->nDeInterlaceHoldingFrameBufferNum > 16)
    {
        if(pVconfig->bThumbnailMode == 0)
        {
            logw("warning: the nDeInterlaceHoldingFrameBufferNum is %d",
            pVconfig->nDeInterlaceHoldingFrameBufferNum);
        }
    }
    if(pVconfig->nDisplayHoldingFrameBufferNum <= 0
       || pVconfig->nDisplayHoldingFrameBufferNum > 16)
    {
        if(pVconfig->bThumbnailMode == 0)
        {
            logw("warning: the nDisplayHoldingFrameBufferNum is %d",
            pVconfig->nDisplayHoldingFrameBufferNum);
        }
    }
    if(pVideoInfo->nHeight >= 3840 || pVideoInfo->nWidth >= 5120)//5K
    {
        pVconfig->bScaleDownEn = 1;
        pVconfig->nHorizonScaleDownRatio = 1;
        pVconfig->nVerticalScaleDownRatio = 1;
    }

    //logi("=====enableAfbc = %d, ENABLE_AFBC = %d", pVconfig->bEnableAfbcFlag, ENABLE_AFBC);
    memcpy(&(p->vconfig), pVconfig, sizeof(VConfig));

    //* create stream buffer.
    nStreamBufferSize = DecideStreamBufferSize(p);
	nStreamBufferSize = 2*1024*1024;
    //* check and fix the configuration for decoder.
    //CheckConfiguration(p);
	p->vconfig.memops = p->memops;
	p->vconfig.veOpsS = p->veops;
//	loge("ji************veops:%lx", p->vconfig.veOpsS);
    //* create video engine.
    p->pVideoEngine = VideoEngineCreate(&(p->vconfig), &(p->videoStreamInfo));
    if(p->pVideoEngine == NULL)
    {
        loge("create video engine fail.");
        return -1;
    }

    SbmConfig mSbmConfig;
    memset(&mSbmConfig, 0, sizeof(SbmConfig));
    mSbmConfig.bVirFlag = p->vconfig.bVirMallocSbm;
    mSbmConfig.memops   = p->memops;
    mSbmConfig.nSbmBufferTotalSize = nStreamBufferSize;
    mSbmConfig.bSecureVideoFlag = p->vconfig.bSecureosEn;
    mSbmConfig.nNaluLength = 4;

    p->pSbm[0] = GetSbmInterfaceStream();

    if(p->pSbm[0] == NULL)
    {
        loge("create stream buffer fail.");
        return -1;
    }

    if(SbmInit(p->pSbm[0], &mSbmConfig) != 0)
    {
        loge(" sbm init failed");
        //SbmDestroy(p->pSbm[0]);
        p->pSbm[0] = NULL;
        return -1;
    }
    VideoEngineSetSbm(p->pVideoEngine, p->pSbm[0], 0);
    p->nSbmNum++;

    if(p->videoStreamInfo.bIs3DStream)
    {
        p->pSbm[1] = GetSbmInterfaceStream();
        if(p->pSbm[1] == NULL)
        {
            loge("create stream buffer fail.");
            return -1;
        }
        if(SbmInit(p->pSbm[1], &mSbmConfig) != 0)
        {
            loge(" sbm init failed");
            SbmDestroy(p->pSbm[1]);
            p->pSbm[1] = NULL;
            return -1;
        }
        VideoEngineSetSbm(p->pVideoEngine, p->pSbm[1], 1);
        p->nSbmNum++;
    }

    return 0;
}

EXPORT_SYMBOL(InitializeVideoDecoder);

void ResetVideoDecoder(VideoDecoder* pDecoder)
{
    int                  i;
    VideoDecoderContext* p;

    logv("ResetVideoDecoder, pDecoder=%p", pDecoder);

    p = (VideoDecoderContext*)pDecoder;
    //Version = VeGetIcVersion();

    //* reset the video engine.
    if(p->pVideoEngine)
    {
        VideoEngineReset(p->pVideoEngine);
    }

    //* flush pictures.
    if(p->nFbmNum == 0 && p->pVideoEngine != NULL)
    {
        p->nFbmNum = VideoEngineGetFbmNum(p->pVideoEngine);
    }

    for(i=0; i<p->nFbmNum; i++)
    {
        if(p->pFbm[i] == NULL && p->pVideoEngine != NULL)
        {
            p->pFbm[i] = VideoEngineGetFbm(p->pVideoEngine, i);
        }

        if(p->pFbm[i] != NULL)
            FbmFlush(p->pFbm[i]);
    }

    //* flush stream data.
    for(i=0; i<p->nSbmNum; i++)
    {
        if(p->pSbm[i] != NULL)
            SbmReset(p->pSbm[i]);
    }

    //* clear the partialStreamDataInfo.
    memset(p->partialStreamDataInfo, 0, sizeof(VideoStreamDataInfo)*2);

    return;
}

int DecoderSetSpecialData(VideoDecoder* pDecoder, void *pArg)
{
    int ret;
    VideoDecoderContext* p;
    DecoderInterface *pDecoderInterface;
    p = (VideoDecoderContext*)pDecoder;

    if(p == NULL ||
        p->pVideoEngine == NULL ||
        p->pVideoEngine->pDecoderInterface == NULL)
    {
        loge(" set decoder special data fail ");
        return VDECODE_RESULT_UNSUPPORTED;
    }

    pDecoderInterface = p->pVideoEngine->pDecoderInterface;
    ret = 0;
    if(pDecoderInterface->SetSpecialData != NULL)
    {
        ret = pDecoderInterface->SetSpecialData(pDecoderInterface, pArg);
    }
    return ret;
}



int DecodeVideoStream(VideoDecoder* pDecoder,
                      int           bEndOfStream,
                      int           bDecodeKeyFrameOnly,
                      int           bDropBFrameIfDelay,
                      int64_t       nCurrentTimeUs)
{
    int                  ret;
    VideoDecoderContext* p;
    int bUpdateProcInfo = 0;

//    logi("DecodeVideoStream, pDecoder=%p, \
//            bEndOfStream=%d, bDropBFrameIfDelay=%d, nCurrentTimeUs=%lld", \
//            pDecoder, bEndOfStream, bDropBFrameIfDelay, nCurrentTimeUs);

   p = (VideoDecoderContext*)pDecoder;

   //eVersion = VeGetIcVersion();

    if(p->pVideoEngine == NULL)
        return VDECODE_RESULT_UNSUPPORTED;

    if(bEndOfStream == 1)
    {
        if(p->pSbm[0])
            SbmSetEos(p->pSbm[0], bEndOfStream);
        if(p->pSbm[1])
            SbmSetEos(p->pSbm[1], bEndOfStream);
    }

    //* also can update proc info in the inter-decode-function
    ret = VideoEngineDecode(p->pVideoEngine,
                            bEndOfStream,
                            bDecodeKeyFrameOnly,
                            bDropBFrameIfDelay,
                            nCurrentTimeUs);

//    dy_log("decode method return %s", strDecodeResult[ret-VDECODE_RESULT_MIN]);

    if(ret == VDECODE_RESULT_FRAME_DECODED ||
       ret == VDECODE_RESULT_KEYFRAME_DECODED)
    {
        p->nDecFrameCount++;
    }

    //* if in thumbnail mode and picture decoded,
    //* reset the video engine to flush the picture out to FBM.
    if(p->vconfig.bThumbnailMode)
    {
        if(ret == VDECODE_RESULT_FRAME_DECODED || ret == VDECODE_RESULT_KEYFRAME_DECODED)
        {
            //VideoEngineReset(p->pVideoEngine);
        }
    }
    else if(bEndOfStream && ret == VDECODE_RESULT_NO_BITSTREAM)
    {
        VideoEngineReset(p->pVideoEngine);
    }

    return ret;
}
EXPORT_SYMBOL(DecodeVideoStream);


int GetVideoStreamInfo(VideoDecoder* pDecoder, VideoStreamInfo* pVideoInfo)
{
    VideoDecoderContext* p;

    logv("GetVideoStreamInfo, pDecoder=%p, pVideoInfo=%p", pDecoder, pVideoInfo);

    p = (VideoDecoderContext*)pDecoder;

    //* set the video stream information.
    pVideoInfo->eCodecFormat   = p->videoStreamInfo.eCodecFormat;
    pVideoInfo->nWidth         = p->videoStreamInfo.nWidth;
    pVideoInfo->nHeight        = p->videoStreamInfo.nHeight;
    pVideoInfo->nFrameRate     = p->videoStreamInfo.nFrameRate;
    pVideoInfo->nFrameDuration = p->videoStreamInfo.nFrameDuration;
    pVideoInfo->nAspectRatio   = p->videoStreamInfo.nAspectRatio;
    pVideoInfo->bIs3DStream    = p->videoStreamInfo.bIs3DStream;

    //* print video stream information.
    {
        logi("Video Stream Information:");
        logi("     codec          = %s",
            strCodecFormat[pVideoInfo->eCodecFormat - VIDEO_CODEC_FORMAT_MIN]);
        logi("     width          = %d pixels", pVideoInfo->nWidth);
        logi("     height         = %d pixels", pVideoInfo->nHeight);
        logi("     frame rate     = %d",        pVideoInfo->nFrameRate);
        logi("     frame duration = %d us",     pVideoInfo->nFrameDuration);
        logi("     aspect ratio   = %d",        pVideoInfo->nAspectRatio);
        logi("     is 3D stream   = %s",        pVideoInfo->bIs3DStream ? "yes" : "no");
    }

    return 0;
}

int RequestVideoStreamBuffer(VideoDecoder* pDecoder,
                             int           nRequireSize,
                             char**        ppBuf,
                             int*          pBufSize,
                             char**        ppRingBuf,
                             int*          pRingBufSize,
                             int           nStreamBufIndex)
{
    char*                pStart;
    char*                pStreamBufEnd;
    char*                pMem;
    int                  nFreeSize;
    SbmInterface*        pSbm;
    VideoDecoderContext* p;

//    logi("RequestVideoStreamBuffer, pDecoder=%p, nRequireSize=%d, nStreamBufIndex=%d",
//            pDecoder, nRequireSize, nStreamBufIndex);

    p = (VideoDecoderContext*)pDecoder;

    *ppBuf        = NULL;
    *ppRingBuf    = NULL;
    *pBufSize     = 0;
    *pRingBufSize = 0;

    pSbm          = p->pSbm[nStreamBufIndex];

    if(pSbm == NULL)
    {
        logw("pSbm of video stream %d is NULL, RequestVideoStreamBuffer fail.", nStreamBufIndex);
        return -1;
    }

    //* sometimes AVI parser will pass empty stream frame to help pts calculation.
    //* in this case give four bytes even the parser does not need.
    if(nRequireSize == 0)
        nRequireSize = 4;

    //* we've filled partial frame data but not added to the SBM before,
    //* we need to calculate the actual buffer pointer by self.
    nRequireSize += p->partialStreamDataInfo[nStreamBufIndex].nLength;
    if(SbmRequestBuffer(pSbm, nRequireSize, &pMem, &nFreeSize) < 0)
    {
        logw("request stream buffer fail,  \
                %d bytes valid data in SBM[%d], total buffer size is %d bytes.",
                SbmStreamDataSize(pSbm),
                nStreamBufIndex,
                SbmBufferSize(pSbm));

        return -1;
    }

    //* check the free buffer is larger than the partial data we filled before.
    if(nFreeSize <= p->partialStreamDataInfo[nStreamBufIndex].nLength)
    {
        logw("require stream buffer get %d bytes, \
                but this buffer has been filled with partial \
                frame data of %d bytes before, nStreamBufIndex=%d.",
                nFreeSize, p->partialStreamDataInfo[nStreamBufIndex].nLength, nStreamBufIndex);

        return -1;
    }

    //* calculate the output buffer pos.
    pStreamBufEnd = (char*)SbmBufferAddress(pSbm) + SbmBufferSize(pSbm);
    pStart        = pMem + p->partialStreamDataInfo[nStreamBufIndex].nLength;
    if(pStart >= pStreamBufEnd)
        pStart -= SbmBufferSize(pSbm);
    nFreeSize -= p->partialStreamDataInfo[nStreamBufIndex].nLength;

    if(pStart + nFreeSize <= pStreamBufEnd) //* check if buffer ring back.
    {
        *ppBuf    = pStart;
        *pBufSize = nFreeSize;
    }
    else
    {
        //* the buffer ring back.
        *ppBuf        = pStreamBufEnd - SbmBufferSize(pSbm);
        *pBufSize     = pStreamBufEnd - pStart;
        *ppRingBuf    = SbmBufferAddress(pSbm);
        *pRingBufSize = nFreeSize - *pBufSize;
        logi("stream buffer %d ring back.", nStreamBufIndex);
    }

    return 0;
}

EXPORT_SYMBOL(RequestVideoStreamBuffer);


void FlushSbmData(VideoDecoderContext* p, SbmInterface*pSbm,
                       VideoStreamDataInfo* pPartialStreamDataInfo)
{
    if((p->vconfig.bVirMallocSbm ==0) &&(pSbm->bUseNewVeMemoryProgram==0))
    {
        //* we need to flush data from cache to memory for the VE hardware module.
        char* pStreamBufEnd = (char*)SbmBufferAddress(pSbm) + SbmBufferSize(pSbm);
        if(pPartialStreamDataInfo->pData + pPartialStreamDataInfo->nLength <= pStreamBufEnd)
        {
            cdc_mem_flush_cache(p->memops, \
                pPartialStreamDataInfo->pData, \
                pPartialStreamDataInfo->nLength);
        }
        else
        {
            //* buffer ring back.
            int nPartialLen = pStreamBufEnd - pPartialStreamDataInfo->pData;
            cdc_mem_flush_cache(p->memops, pPartialStreamDataInfo->pData, nPartialLen);
            cdc_mem_flush_cache(p->memops, \
                SbmBufferAddress(pSbm), \
                pPartialStreamDataInfo->nLength - nPartialLen);

        }
    }
}
int SubmitVideoStreamData(VideoDecoder*        pDecoder,
                          VideoStreamDataInfo* pDataInfo,
                          int                  nStreamBufIndex)
{
    SbmInterface*                 pSbm;
    VideoDecoderContext* p;
    VideoStreamDataInfo* pPartialStreamDataInfo;
    struct mem_interface *_memops;
    char mask[6] = "awsp";

    p = (VideoDecoderContext*)pDecoder;

//    dy_log("Submit, pDecoder=%p, pDataInfo=%p, Index=%d, len = %d, pts = %lld",
//            pDecoder, pDataInfo, nStreamBufIndex,
//            pDataInfo->nLength, (long long int)pDataInfo->nPts);

    _memops = p->memops;

    pSbm = p->pSbm[nStreamBufIndex];

    if(pSbm == NULL)
    {
        logw("pSbm of video stream %d is NULL, SubmitVideoStreamData fail.", nStreamBufIndex);
        return -1;
    }

    pPartialStreamDataInfo = &p->partialStreamDataInfo[nStreamBufIndex];

    //* chech wheter a new stream frame.
    if(pDataInfo->bIsFirstPart)
    {
        if(pPartialStreamDataInfo->nLength != 0)    //* last frame is not complete yet.
        {
            FlushSbmData(p, pSbm, pPartialStreamDataInfo);
            SbmAddStream(pSbm, pPartialStreamDataInfo);
        }

        //* set the data address and pts.
        pPartialStreamDataInfo->pData        = pDataInfo->pData;
        pPartialStreamDataInfo->nLength      = pDataInfo->nLength;
        pPartialStreamDataInfo->nPts         = pDataInfo->nPts;
        pPartialStreamDataInfo->nPcr         = pDataInfo->nPcr;
        pPartialStreamDataInfo->bIsFirstPart = pDataInfo->bIsFirstPart;
        pPartialStreamDataInfo->bIsLastPart  = 0;
        pPartialStreamDataInfo->bVideoInfoFlag = pDataInfo->bVideoInfoFlag;
        pPartialStreamDataInfo->pVideoInfo     = pDataInfo->pVideoInfo;
    }
    else
    {
        if(pPartialStreamDataInfo->pData == NULL)
        {
            logd("pPartialStreamDataInfo->pData == NULL");
            pPartialStreamDataInfo->pData = pDataInfo->pData;
        }
        pPartialStreamDataInfo->nLength += pDataInfo->nLength;
        if(pPartialStreamDataInfo->nPts == -1 && pDataInfo->nPts != -1)
            pPartialStreamDataInfo->nPts = pDataInfo->nPts;
    }

    //* check whether a stream frame complete.
    if(pDataInfo->bIsLastPart)
    {
        if(pPartialStreamDataInfo->pData != NULL && pPartialStreamDataInfo->nLength != 0)
        {
            FlushSbmData(p, pSbm, pPartialStreamDataInfo);
        }
        else
        {
            //* maybe it is a empty frame for MPEG4 decoder from AVI parser.
            logw("empty stream data frame submitted, pData=%p, nLength=%d",
                pPartialStreamDataInfo->pData, pPartialStreamDataInfo->nLength);
        }

        SbmAddStream(pSbm, pPartialStreamDataInfo);

        //* clear status of stream data info.
        pPartialStreamDataInfo->pData        = NULL;
        pPartialStreamDataInfo->nLength      = 0;
        pPartialStreamDataInfo->nPts         = -1;
        pPartialStreamDataInfo->nPcr         = -1;
        pPartialStreamDataInfo->bIsLastPart  = 0;
        pPartialStreamDataInfo->bIsFirstPart = 0;
        pPartialStreamDataInfo->bVideoInfoFlag = 0;
        pPartialStreamDataInfo->pVideoInfo     = NULL;
    }

    return 0;
}
EXPORT_SYMBOL(SubmitVideoStreamData);

int VideoStreamBufferSize(VideoDecoder* pDecoder, int nStreamBufIndex)
{
    VideoDecoderContext* p = (VideoDecoderContext*)pDecoder;

    logi("VideoStreamBufferSize, nStreamBufIndex=%d", nStreamBufIndex);

    if(p->pSbm[nStreamBufIndex] == NULL)
        return 0;

    return SbmBufferSize(p->pSbm[nStreamBufIndex]);
}

int VideoStreamDataSize(VideoDecoder* pDecoder, int nStreamBufIndex)
{
    VideoDecoderContext* p = (VideoDecoderContext*)pDecoder;

    logi("VideoStreamDataSize, nStreamBufIndex=%d", nStreamBufIndex);

    if(p->pSbm[nStreamBufIndex] == NULL)
        return 0;

    return SbmStreamDataSize(p->pSbm[nStreamBufIndex]);
}

int VideoStreamFrameNum(VideoDecoder* pDecoder, int nStreamBufIndex)
{
    VideoDecoderContext* p = (VideoDecoderContext*)pDecoder;

    logi("VideoStreamFrameNum, nStreamBufIndex=%d", nStreamBufIndex);

    if(p->pSbm[nStreamBufIndex] == NULL)
        return 0;

    return SbmStreamFrameNum(p->pSbm[nStreamBufIndex]);
}

void* VideoStreamDataInfoPointer(VideoDecoder* pDecoder, int nStreamBufIndex)
{
    VideoDecoderContext* p = (VideoDecoderContext*)pDecoder;

    logi("VideoStreamDataSize, nStreamBufIndex=%d", nStreamBufIndex);

    if(p->pSbm[nStreamBufIndex] == NULL)
        return NULL;

    return SbmBufferDataInfo(p->pSbm[nStreamBufIndex]);
}

VideoPicture* RequestPicture(VideoDecoder* pDecoder, int nStreamIndex)
{
    VideoDecoderContext* p;
    Fbm*                 pFbm;
    VideoPicture*        pPicture;

//    logi("RequestPicture, nStreamIndex=%d", nStreamIndex);

    p = (VideoDecoderContext*)pDecoder;

    pFbm = p->pFbm[nStreamIndex];
    if(pFbm == NULL)
    {
        //* get FBM handle from video engine.
        if(p->pVideoEngine != NULL)
            pFbm = p->pFbm[nStreamIndex] = VideoEngineGetFbm(p->pVideoEngine, nStreamIndex);

        if(pFbm == NULL)
        {
            logi("Fbm module of video stream %d not create yet, \
                RequestPicture fail.", nStreamIndex);
            return NULL;
        }
    }

    //* request picture.
    pPicture = FbmRequestPicture(pFbm);

    if(pPicture != NULL)
    {
        if(pPicture->colour_primaries == 9)
        {
            pPicture->matrix_coeffs = VIDEO_MATRIX_COEFFS_BT2020;
        }

        #if(CHECK_METADATA_INFO)
        printfMetaDataInfo((struct sunxi_metadata*)pPicture->pMetaData);
        #endif

//        logv("**** color_primaries = %d, matrix_coeffs = %d",
//             pPicture->colour_primaries,pPicture->matrix_coeffs);
        //* set stream index to the picture, it will be useful when picture returned.
        pPicture->nStreamIndex = nStreamIndex;

        //* update video stream information.
        UpdateVideoStreamInfo(p, pPicture);
//        loge("picture w: %d,  h: %d, addr: %p, top: %d, \
//                bottom: %d, left: %d, right: %d, stride: %d", \
//                pPicture->nWidth, pPicture->nHeight, pPicture,
//                pPicture->nTopOffset, pPicture->nBottomOffset,
//                pPicture->nLeftOffset, pPicture->nRightOffset, pPicture->nLineStride);
        //DebugSavePicture(p, pPicture);
        
        //CdcMemFlushCache((struct ScMemOpsS *)(p->memops), pPicture->pData0, pPicture->nWidth*pPicture->nHeight*3/2);
    }

    return pPicture;
}
EXPORT_SYMBOL(RequestPicture);


int ReturnPicture(VideoDecoder* pDecoder, VideoPicture* pPicture)
{
    VideoDecoderContext* p;
    Fbm*                 pFbm;
    int                  nStreamIndex;
    int                  ret;

//    logi("ReturnPicture, pPicture=%p", pPicture);

    p = (VideoDecoderContext*)pDecoder;

    if(pPicture == NULL)
        return 0;
    nStreamIndex = pPicture->nStreamIndex;
    if(nStreamIndex < 0 || nStreamIndex > 2)
    {
        loge("invalid stream index(%d), pPicture->nStreamIndex must had been \
                changed by someone incorrectly, ReturnPicture fail.", nStreamIndex);
        return -1;
    }

    pFbm = p->pFbm[nStreamIndex];
    if(pFbm == NULL)
    {
        logw("pFbm is NULL when returning a picture, ReturnPicture fail.");
        return -1;
    }

    ret = FbmReturnPicture(pFbm, pPicture);
    if(ret != 0)
    {
        logw("FbmReturnPicture return fail,\
            it means the picture being returned it not one of this FBM.");
    }

    return ret;
}
EXPORT_SYMBOL(ReturnPicture);


VideoPicture* NextPictureInfo(VideoDecoder* pDecoder, int nStreamIndex)
{
    VideoDecoderContext* p;
    Fbm*                 pFbm;
    VideoPicture*        pPicture;

//    logi("RequestPicture, nStreamIndex=%d", nStreamIndex);

    p = (VideoDecoderContext*)pDecoder;

    pFbm = p->pFbm[nStreamIndex];
    if(pFbm == NULL)
    {
        //* get FBM handle from video engine.
        if(p->pVideoEngine != NULL)
            pFbm = p->pFbm[nStreamIndex] = VideoEngineGetFbm(p->pVideoEngine, nStreamIndex);

        logi("Fbm module of video stream %d not create yet, NextPictureInfo() fail.", nStreamIndex);
        return NULL;
    }

    //* request picture.
    pPicture = FbmNextPictureInfo(pFbm);

    if(pPicture != NULL)
    {
        //* set stream index to the picture, it will be useful when picture returned.
        pPicture->nStreamIndex = nStreamIndex;

        //* update video stream information.
        UpdateVideoStreamInfo(p, pPicture);
    }

    return pPicture;
}

int TotalPictureBufferNum(VideoDecoder* pDecoder, int nStreamIndex)
{
    VideoDecoderContext* p;
    Fbm*                 pFbm;

//    logv("TotalPictureBufferNum, nStreamIndex=%d", nStreamIndex);

    p    = (VideoDecoderContext*)pDecoder;
    pFbm = p->pFbm[nStreamIndex];
    if(pFbm == NULL)
    {
        //* get FBM handle from video engine.
        if(p->pVideoEngine != NULL)
            pFbm = p->pFbm[nStreamIndex] = VideoEngineGetFbm(p->pVideoEngine, nStreamIndex);

        logi("Fbm module of video stream %d not create yet, \
            TotalPictureBufferNum() fail.", nStreamIndex);
        return 0;
    }

    return FbmTotalBufferNum(pFbm);
}

int EmptyPictureBufferNum(VideoDecoder* pDecoder, int nStreamIndex)
{
    VideoDecoderContext* p;
    Fbm*                 pFbm;

//    logv("EmptyPictureBufferNum, nStreamIndex=%d", nStreamIndex);

    p    = (VideoDecoderContext*)pDecoder;
    pFbm = p->pFbm[nStreamIndex];
    if(pFbm == NULL)
    {
        //* get FBM handle from video engine.
        if(p->pVideoEngine != NULL)
            pFbm = p->pFbm[nStreamIndex] = VideoEngineGetFbm(p->pVideoEngine, nStreamIndex);

        logi("Fbm module of video stream %d not create yet, \
            EmptyPictureBufferNum() fail.", nStreamIndex);
        return 0;
    }

    return FbmEmptyBufferNum(pFbm);
}

int ValidPictureNum(VideoDecoder* pDecoder, int nStreamIndex)
{
    VideoDecoderContext* p;
    Fbm*                 pFbm;

//    logv("ValidPictureNum, nStreamIndex=%d", nStreamIndex);

    p    = (VideoDecoderContext*)pDecoder;
    pFbm = p->pFbm[nStreamIndex];
    if(pFbm == NULL)
    {
        //* get FBM handle from video engine.
        if(p->pVideoEngine != NULL)
            pFbm = p->pFbm[nStreamIndex] = VideoEngineGetFbm(p->pVideoEngine, nStreamIndex);

        logi("Fbm module of video stream %d not create yet, ValidPictureNum() fail.", nStreamIndex);
        return 0;
    }

    return FbmValidPictureNum(pFbm);
}

//***********************************************//
//***********************************************//
//***added by xyliu at 2015-12-23****************//
/*if you want to scale the output video frame, but you donot know the real size of video,
 cannot set the scale params (nHorizonScaleDownRatio,nVerticalScaleDownRatio,
 bRotationEn), so you can set value to ExtraVidScaleInfo
*/
int ConfigExtraScaleInfo(VideoDecoder* pDecoder, int nWidthTh,
            int nHeightTh, int nHorizonScaleRatio, int nVerticalScaleRatio)
{
    VideoDecoderContext* p;
    int ret;
    DecoderInterface *pDecoderInterface;
    p = (VideoDecoderContext*)pDecoder;

    logd("*************config ConfigExtraScaleInfo\n");

    if(p == NULL ||p->pVideoEngine == NULL ||
            p->pVideoEngine->pDecoderInterface == NULL)
    {
        loge(" set decoder extra scale info fail ");
        return VDECODE_RESULT_UNSUPPORTED;
    }
    pDecoderInterface = p->pVideoEngine->pDecoderInterface;
    ret = 0;
    if(pDecoderInterface->SetExtraScaleInfo != NULL)
    {
        ret = pDecoderInterface->SetExtraScaleInfo(pDecoderInterface, \
                    nWidthTh, nHeightTh,nHorizonScaleRatio, nVerticalScaleRatio);
    }
    return ret;
}

int ReopenVideoEngine(VideoDecoder* pDecoder, VConfig* pVConfig, VideoStreamInfo* pStreamInfo)
{
    int                  i;
    VideoDecoderContext* p;
    int                  ret = 0;

    logv("ReopenVideoEngine");

    p = (VideoDecoderContext*)pDecoder;
    if(p->pVideoEngine == NULL)
    {
        logw("video decoder is not initialized yet, ReopenVideoEngine() fail.");
        return -1;
    }
    //* create a video engine again,
    //* before that, we should set vconfig and videoStreamInfo again.
    if(p->videoStreamInfo.pCodecSpecificData != NULL)
    {
        kfree(p->videoStreamInfo.pCodecSpecificData);
        p->videoStreamInfo.pCodecSpecificData = NULL;
    }

    ret = VideoEngineReopen(p->pVideoEngine,pVConfig, pStreamInfo);
    if(ret < 0)
    {
        loge("VideoEngineReopen result is %d\n", ret);
        return -1;
    }

    p->pFbm[0] = p->pFbm[1] = NULL;
    p->nFbmNum = 0;

    memcpy(&p->vconfig,pVConfig,sizeof(VConfig));
    memcpy(&p->videoStreamInfo,pStreamInfo,sizeof(VideoStreamInfo));
    pStreamInfo->bReOpenEngine = 1;

    if(pStreamInfo->pCodecSpecificData)
    {
        p->videoStreamInfo.pCodecSpecificData = (char*)kmalloc(pStreamInfo->nCodecSpecificDataLen, GFP_KERNEL | __GFP_ZERO);
        if(p->videoStreamInfo.pCodecSpecificData == NULL)
        {
            loge("malloc video codec specific data fail!");
            return -1;
        }
        memcpy(p->videoStreamInfo.pCodecSpecificData,
               pStreamInfo->pCodecSpecificData,
               pStreamInfo->nCodecSpecificDataLen);
        p->videoStreamInfo.nCodecSpecificDataLen = pStreamInfo->nCodecSpecificDataLen;
    }
    else
    {
       p->videoStreamInfo.pCodecSpecificData    = NULL;
       p->videoStreamInfo.nCodecSpecificDataLen = 0;
    }

    //* set sbm module to the video engine.
    for(i=0; i<p->nSbmNum; i++)
        VideoEngineSetSbm(p->pVideoEngine, p->pSbm[i], i);

    return 0;
}

#define BUFFER_SIZE_480P   340000  //(720x480)
#define BUFFER_SIZE_720P   770000  //(1080x720)
#define BUFFER_SIZE_1080P  2000000 //(1920x1080)
#define BUFFER_SIZE_2K     4500000 //(2560x1920)
#define BUFFER_SIZE_4K     7000000 //(3840x2160)

static int DecideStreamBufferSize(VideoDecoderContext* p)
{
    int nPixels;
    int eCodecFormat;
    int nBufferSize = 8*1024*1024;

    //* we decide stream buffer size by resolution and codec format.
    nPixels = p->videoStreamInfo.nWidth * p->videoStreamInfo.nHeight;
    eCodecFormat = p->videoStreamInfo.eCodecFormat;

    //* for skia create sbm
    if((p->vconfig.eOutputPixelFormat == PIXEL_FORMAT_RGBA)
        && (eCodecFormat == VIDEO_CODEC_FORMAT_MJPEG)
        && p->vconfig.nVbvBufferSize)
    {
        nBufferSize = p->vconfig.nVbvBufferSize;
        logd("nBufferSize=%d", nBufferSize);
        return nBufferSize;
    }

    //* if resolution is unknown, treat it as full HD source.
    if(nPixels == 0)
        nPixels = BUFFER_SIZE_1080P;

    if(nPixels < BUFFER_SIZE_480P)
        nBufferSize = 2*1024*1024;
    else if (nPixels < BUFFER_SIZE_720P)
        nBufferSize = 4*1024*1024;
    else if(nPixels < BUFFER_SIZE_1080P)
        nBufferSize = 6*1024*1024;
    else if(nPixels < BUFFER_SIZE_2K)
        nBufferSize = 8*1024*1024;
    else if(nPixels < BUFFER_SIZE_4K)
        nBufferSize = 16*1024*1024;
    else
        nBufferSize = 32*1024*1024;

    if(eCodecFormat == VIDEO_CODEC_FORMAT_MJPEG ||
       eCodecFormat == VIDEO_CODEC_FORMAT_MPEG1 ||
       eCodecFormat == VIDEO_CODEC_FORMAT_MPEG2)
    {
        nBufferSize += 2*1024*1024; //* for old codec format, compress rate is low.
    }
    if(eCodecFormat == VIDEO_CODEC_FORMAT_RX)
    {
        nBufferSize = 16*1024*1024;
    }

    if(nPixels >= BUFFER_SIZE_4K && nBufferSize<=32*1024*1024)
    {
        nBufferSize = 32*1024*1024;
    }

    //* decrease the size as the parser is simple on ipc/cdr-chip,such as V3 V5
#ifdef CONF_SIMPLE_PARSER_CASE
    if(nPixels >= BUFFER_SIZE_4K)
        nBufferSize = 12*1024*1024;
#endif

    if((p->vconfig.nVbvBufferSize>=1*1024*1024) && (p->vconfig.nVbvBufferSize<=32*1024*1024))
    {
        nBufferSize = p->vconfig.nVbvBufferSize > nBufferSize ? p->vconfig.nVbvBufferSize : nBufferSize;
        logd("nBufferSize=%d", nBufferSize);
        return nBufferSize;
    }

    logd("nBufferSize=%d", nBufferSize);
    return nBufferSize;
}


static void UpdateVideoStreamInfo(VideoDecoderContext* p, VideoPicture* pPicture)
{
    //* update video resolution.
    if(p->vconfig.bRotationEn   == 0 ||
       p->vconfig.nRotateDegree == 0 ||
       p->vconfig.nRotateDegree == 3)
    {
        //* the picture was not rotated or rotated by 180 degree.
        if(p->vconfig.bScaleDownEn == 0)
        {
            //* the picture was not scaled.
            if(p->videoStreamInfo.nWidth  != pPicture->nWidth ||
               p->videoStreamInfo.nHeight != pPicture->nHeight)
            {
                p->videoStreamInfo.nWidth  = pPicture->nWidth;
                p->videoStreamInfo.nHeight = pPicture->nHeight;
            }
        }
        else
        {
            //* the picture was scaled.
            if(p->videoStreamInfo.nWidth  != \
                (pPicture->nWidth<<p->vconfig.nHorizonScaleDownRatio) ||
               p->videoStreamInfo.nHeight != \
               (pPicture->nHeight<<p->vconfig.nVerticalScaleDownRatio))
            {
                p->videoStreamInfo.nWidth  = \
                    pPicture->nWidth << p->vconfig.nHorizonScaleDownRatio;
                p->videoStreamInfo.nHeight = \
                    pPicture->nHeight << p->vconfig.nVerticalScaleDownRatio;
            }
        }
    }
    else
    {
        //* the picture was rotated by 90 or 270 degree.
        if(p->vconfig.bScaleDownEn == 0)
        {
            //* the picture was not scaled.
            if(p->videoStreamInfo.nWidth  != pPicture->nHeight ||
               p->videoStreamInfo.nHeight != pPicture->nWidth)
            {
                p->videoStreamInfo.nWidth  = pPicture->nHeight;
                p->videoStreamInfo.nHeight = pPicture->nWidth;
            }
        }
        else
        {
            //* the picture was scaled.
            if(p->videoStreamInfo.nWidth  != \
                (pPicture->nHeight<<p->vconfig.nHorizonScaleDownRatio) ||
               p->videoStreamInfo.nHeight != \
               (pPicture->nWidth<<p->vconfig.nVerticalScaleDownRatio))
            {
                p->videoStreamInfo.nWidth  = \
                    pPicture->nHeight << p->vconfig.nHorizonScaleDownRatio;
                p->videoStreamInfo.nHeight = \
                    pPicture->nWidth << p->vconfig.nVerticalScaleDownRatio;
            }
        }
    }

    //* update frame rate.
    if(p->videoStreamInfo.nFrameRate != pPicture->nFrameRate && pPicture->nFrameRate != 0)
        p->videoStreamInfo.nFrameRate = pPicture->nFrameRate;

    return;
}

//*****************************************************************************************//
//***added new interface function for new display structure********************************//
//******************************************************************************************//
FbmBufInfo* GetVideoFbmBufInfo(VideoDecoder* pDecoder)
{
     VideoDecoderContext* p;

     p = (VideoDecoderContext*)pDecoder;

     if(p == NULL)
     {
         return NULL;
     }
     if(p->pVideoEngine != NULL)
     {
         int nStreamIndex = 0;
         if(VideoEngineGetFbm(p->pVideoEngine, nStreamIndex) == NULL)
         {
             return NULL;
         }
         return FbmGetVideoBufferInfo(&(p->pVideoEngine->fbmInfo));
     }
     return NULL;
}

VideoPicture* SetVideoFbmBufAddress(VideoDecoder* pDecoder,
                                        VideoPicture* pVideoPicture, int bForbidUseFlag)
{
    VideoDecoderContext* p;
    p = (VideoDecoderContext*)pDecoder;

    if(p == NULL)
    {
        return NULL;
    }

    if(p->vconfig.bGpuBufValid)
    {

        pVideoPicture->phyYBufAddr = 0;
        pVideoPicture->phyYBufAddr -= 0;
        pVideoPicture->phyCBufAddr = pVideoPicture->phyYBufAddr +
                                            (pVideoPicture->nHeight * pVideoPicture->nLineStride);
        if(pVideoPicture->nHeight * pVideoPicture->nLineStride % 1024 != 0)
        {
            loge("phyCBufAddr is not aligned to 1024");
        }
    }
    if(p->pVideoEngine != NULL)
    {
        return FbmSetFbmBufAddress(&(p->pVideoEngine->fbmInfo), pVideoPicture, bForbidUseFlag);
    }
    return NULL;
}

int SetVideoFbmBufRelease(VideoDecoder* pDecoder)
{
    VideoDecoderContext* p;
    p = (VideoDecoderContext*)pDecoder;

    if(p == NULL)
    {
        return -1;
    }
    if(p->pVideoEngine != NULL)
    {
        return FbmSetNewDispRelease(&(p->pVideoEngine->fbmInfo));
    }
    return -1;
}

VideoPicture* RequestReleasePicture(VideoDecoder* pDecoder)
{
    VideoDecoderContext* p;
    p = (VideoDecoderContext*)pDecoder;
    VideoPicture* pVpicture = NULL;

    if(p == NULL)
    {
        return NULL;
    }
    if(p->pVideoEngine != NULL)
    {
         //return FbmRequestReleasePicture(&(p->pVideoEngine->fbmInfo));
        pVpicture = FbmRequestReleasePicture(&(p->pVideoEngine->fbmInfo));
        
        return pVpicture;
    }
    return NULL;
}

VideoPicture*  ReturnRelasePicture(VideoDecoder* pDecoder, \
        VideoPicture* pVpicture, int bForbidUseFlag)
{
    VideoDecoderContext* p;
    p = (VideoDecoderContext*)pDecoder;

    if(p == NULL)
    {
        return NULL;
    }
    if(p->pVideoEngine != NULL)
    {
        
        return FbmReturnReleasePicture(&(p->pVideoEngine->fbmInfo), pVpicture, bForbidUseFlag);
    }
    return NULL;
}
