
/*
* Copyright (c) 2008-2016 Allwinner Technology Co. Ltd.
* All rights reserved.
*
* File : mjpeg_dec.h
* Description :
* History :
*   Author  : xyliu <xyliu@allwinnertech.com>
*   Date    : 2016/04/13
*   Comment :
*
*
*/

#ifndef JPEG_DEC_LIB_H
#define JPEG_DEC_LIB_H

#include "mjpeg.h"
#include "mjpeg_context.h"

#ifdef __cplusplus
extern "C" {
#endif

    //JPEG Define

    #define MAX(a,b)    ((a) > (b) ? (a) : (b))
    #define MIN(a,b)    ((a) < (b) ? (a) : (b))

    typedef enum JPEG_MARKER    /* JPEG marker codes */
    {
        M_SOF0  = (unsigned)0xc0,
        M_SOF1  = (unsigned)0xc1,
        M_SOF2  = (unsigned)0xc2,
        M_SOF3  = (unsigned)0xc3,

        M_SOF5  = (unsigned)0xc5,
        M_SOF6  = (unsigned)0xc6,
        M_SOF7  = (unsigned)0xc7,

        M_JPG   = (unsigned)0xc8,
        M_SOF9  = (unsigned)0xc9,
        M_SOF10 = (unsigned)0xca,
        M_SOF11 = (unsigned)0xcb,

        M_SOF13 = (unsigned)0xcd,
        M_SOF14 = (unsigned)0xce,
        M_SOF15 = (unsigned)0xcf,

        M_DHT   = (unsigned)0xc4,

        M_DAC   = (unsigned)0xcc,

        M_RST0  = (unsigned)0xd0,
        M_RST1  = (unsigned)0xd1,
        M_RST2  = (unsigned)0xd2,
        M_RST3  = (unsigned)0xd3,
        M_RST4  = (unsigned)0xd4,
        M_RST5  = (unsigned)0xd5,
        M_RST6  = (unsigned)0xd6,
        M_RST7  = (unsigned)0xd7,

        M_SOI   = (unsigned)0xd8,
        M_EOI   = (unsigned)0xd9,
        M_SOS   = (unsigned)0xda,
        M_DQT   = (unsigned)0xdb,
        M_DNL   = (unsigned)0xdc,
        M_DRI   = (unsigned)0xdd,
        M_DHP   = (unsigned)0xde,
        M_EXP   = (unsigned)0xdf,

        M_APP0  = (unsigned)0xe0,
        M_APP1  = (unsigned)0xe1,
        M_APP2  = (unsigned)0xe2,
        M_APP3  = (unsigned)0xe3,
        M_APP4  = (unsigned)0xe4,
        M_APP5  = (unsigned)0xe5,
        M_APP6  = (unsigned)0xe6,
        M_APP7  = (unsigned)0xe7,
        M_APP8  = (unsigned)0xe8,
        M_APP9  = (unsigned)0xe9,
        M_APP10 = (unsigned)0xea,
        M_APP11 = (unsigned)0xeb,
        M_APP12 = (unsigned)0xec,
        M_APP13 = (unsigned)0xed,
        M_APP14 = (unsigned)0xee,
        M_APP15 = (unsigned)0xef,

        M_JPG0  = (unsigned)0xf0,
        M_JPG13 = (unsigned)0xfd,
        M_COM   = (unsigned)0xfe,

        M_TEM   = (unsigned)0x01,

        M_ERROR = (unsigned)0x100
    }JPEG_MARKER;

    typedef enum MJPEG_DEC_STEP
    {
        MJPEG_DEC_INIT           = 0,
        MJPEG_DEC_GET_FBM_BUFFER = 1,
        MJPEG_DEC_GET_STREAM     = 2,
        MJPEG_DEC_STREAM_DATA    = 3,
        MJPEG_DEC_PROCESS_ROTATE = 4,
        MJPEG_DEC_PROCESS_RESULT = 5,
        MJPEG_DEC_Step_
    }MjpegDecStep;

    typedef enum JPEG_PARSE_FORMAT
    {
        JPEG420 = 0,
        JPEG411 ,
        JPEG422    ,
        JPEG444    ,
        JPEG422T,
        JPEG400    ,
        JPEGERR
    }JpegPaserFormat;

    typedef enum JPEG_PARSER_STATE
    {
        JPEG_MARKER_ERROR   =0,
        JPEG_FORMAT_UNSUPPORT,
        JPEG_PARSER_OK
    }JpegParserState;

    #define MAX_COMPONENTS    3
    #define BITS_IN_JSAMPLE   8
    #define DCTSIZE2          64
    #define NUM_QUANT_TBLS      4
    #define NUM_HUFF_TBLS     2
    #define MAX_COMPS_IN_SCAN 4

    typedef struct JPEG_HUFF_TABLE
    {
        u16 startCode[16];
        u8  offset[16];
        u8  symbol[256];
    }JHuffTable;

    typedef struct JPEG_COMP_INFO
    {
        s32   componentId;
        s32   componentIndex;
        s16   hSampFactor;
        s16   vSampFactor;
        s32   quantTblNo;

        s32   dcTblNo;
        s32   acTblNo;
    }JpegComponentInfo;

    typedef enum DISP_PIXEL_FORMAT
    {
        DISP_FORMAT_YU12 = 2,
        DISP_FORMAT_YV12 = 3,
        DISP_FORMAT_NV12 = 4,
        DISP_FORMAT_NV21 = 5,
    }eDispPixelFormat;

    typedef struct JPEG_DECODER
    {
        u8                  nDecStramIndex;
        u8                  nDecStep;

        SbmInterface*       pSbm;
        u8*                 pVbvBase;
        u32                 nVbvSize;
        Fbm*                pFbm;
        Fbm*                pFbmScaleDown;

        u8*                    data;
        u32                    nFrameSize;

        s32                    nWidth;
        s32                 nHeight;
        s32                 nNumComponents;
        s32                 nCompsInScan;
        s32                 nNextRestartNum;
        s32                 nRestartInterval;
        JpegComponentInfo   curCompInfo[MAX_COMPONENTS];

        u8                  sosNbBlocks[MAX_COMPS_IN_SCAN];
        u8                  sosHCount[MAX_COMPS_IN_SCAN];
        u8                  sosVCount[MAX_COMPS_IN_SCAN];
        u8                  seqCompId[MAX_COMPS_IN_SCAN];

        s32                 maxHSampFactor;
        s32                 maxVSampFactor;

        u16                 eoiReached;
        s32                 jpegFormat;
        u16                 mcuWidth;
        u16                 mcuHeight;
        s32                 isHuffSramA;

        u16                 QTab[4][DCTSIZE2];    /* quantization step for each coefficient */
        JHuffTable             dcHuffTbl[NUM_HUFF_TBLS];
        JHuffTable             acHuffTbl[NUM_HUFF_TBLS];

        u8                  unreadMarker;
        u8                  sawSOI;
        u8                  sawSOF;

        s32                 bufferFillOver;
        s64                  lastPts;

        u16                 hasDht;
        u16                 hasConfigDht;

        s8                  bodyfrmFlag;
        // 0-first frame, !0-not the first frame,so call body frame
        s8                    syncstatFlag;
        //if set 1, indicate play mode change, need sync with parser,

        //and vdrv will clear frame buffer and output frame queue.

        //It is used for FF, RR, NLY(normal play) switch

        s32                    enRotationOut;
        s32                    enScaleDownOut;
        s32                 verticalScaleRatio;
        s32                 horizontalScaleRatio;
        s32                 rotationDegree;

        s64                 nPts;
        u8                  bHwForce420Out;
        // use the hardware to convert the data into yuv420

        //u8                  bRefMbDispPlanner;
        //u8                  bRefPlannerDispPlanner;
        VideoPicture*       pRefPicture;
        VideoPicture*       pReconPicture;
        VideoPicture*       pDispPicture;
        u32                 nRefPicLumaAddr;
        u32                 nRefPicChromaAddr;
        u32                 nReconPicLumaAddr;
        u32                 nReconPicChromaAddr;
        u64                 nVeVersion;
        s32                 bYv12OutFlag;
        s32                 bNv21OutFlag;
        u32                 nDispCBufferSize;
        u32                 nRefCBufferSize;
        u32                 nRefYLineStride;
        u32                 nRefCLineStride;
        u32                 nDispYLineStride;
        u32                 nDispCLineStride;

        u16*                pHuffmanTableSramBuf;
        s32                 nLeftOffset;
        s32                 nRightOffset;
        s32                 nTopOffset;
        s32                 nBottomOffset;

        u32                 nDecFrameCount;
        eDispPixelFormat  eDispFormat;
        u8      bRefDispSameFormat;

    }JpegDec;

    s32 JpegDecoderMain(MjpegDecodeContext* pMjpegContext,  JpegDec* jpgctx);
    s32 JpegDecodeRaw(JpegDec* p);
	s32 RotatMjpegByHardware(MjpegDecodeContext* pMjpegContex, u32 nInLumaBufAddr,
    	u32 nInChromaBufAddr, u32 nOutLumaBufAddr, u32 nOutChromaBufAddr);
	void MjpegTransformPlanner422To420(MjpegDecodeContext* pMjpegContext,JpegDec*pMjpegDec, u8*Y);
	void MjpegSetDispParams(MjpegDecodeContext* pMjpegContext, JpegDec* mJpegDec, VideoPicture* pCurPicture);
	s32 RotateMjpegBySoft(u8* JpegLumaSrcStartAddr,u8* JpegLumaOutStartAddr,s32 rotateAngle,
    	s32 width, s32 height, MjpegDecodeContext* pMjpegContext);
#ifdef __cplusplus
}
#endif

#endif

