/*
 * =====================================================================================
 *
 *       Filename:  aw_ve.c
 *
 *    Description:
 *
 *        Version:  1.0
 *        Created:  2020-04-16
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  jilinglin,<jilinglin@allwinnertech.com>
 *        Company:
 *
 * =====================================================================================
 */
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/string.h>

#include "cedar_ve_out.h"
#include "log.h"
#include "top_reg_def.h"
#include "ve_interface.h"
#include "ve_policy.h"

#define REG_GROUP_VETOP_OFFSET 0x0

#define REG_GROUP_MPEG_OFFSET 0x100
#define REG_GROUP_H264_OFFSET 0x200
#define REG_GROUP_AVS_OFFSET 0x200
#define REG_GROUP_VC1_OFFSET 0x300
#define REG_GROUP_RV_OFFSET 0x400
#define REG_GROUP_H265_OFFSET 0x500
#define REG_GROUP_AVS2_OFFSET 0x500
#define REG_GROUP_JPEG_OFFSET 0xe00

#define TOP_MODE_SELECT_OFFSET 0x00
#define TOP_VE_RESET_OFFSET 0x04
#define TOP_VE_JPEG_RESET_OFFSET 0x2c

struct ve_aw_ctx {
	struct ve_interface interface;
	char* reg_base;
	unsigned long long ve_version;
	unsigned int phy_offset;
	struct ve_config config;
	int ve_list_num;
	char soc_str[20];
};

static int get_ve_hardware_version(struct ve_aw_ctx* context)
{
	volatile unsigned int value;
	unsigned int read_count = 0;

READ_IC_VERSION:
	value = *((unsigned int*)(context->reg_base + 0xf0));
	printk("value = %d\n", value);
	if (value == 0 || value == 0x16670040) {
		volatile unsigned int value_enc_ip;
		volatile unsigned int value_dec_ip;
		unsigned long long dec_enc_ip;

		value_dec_ip = *((unsigned int*)(context->reg_base + 0xe0));
		logv("the get dec_ip_version:%08x, reg_base = 0x%x\n", value_dec_ip,
			 context->reg_base);

		if (value_dec_ip == 0) {
			logw(
				"can not get the ve version ,both 0xf0 and 0xe0 is 0, try to "
				"read again\n");
			printk(
				"can not get the ve version ,both 0xf0 and 0xe0 is 0, try to "
				"read again\n");
			context->ve_version = 0;

			// we should read the ic version again when other process's codec is
			// reseting the ve
			read_count++;
			if (read_count < 4) {
				mdelay(1);
				goto READ_IC_VERSION;
			} else
				return -1;
		}

		value_enc_ip = *((unsigned int*)(context->reg_base + 0xe4));
		logv("the get enc_ip_version:%08x\n", value_enc_ip);
		printk("the get enc_ip_version:%08x\n", value_enc_ip);
		if (value_enc_ip == 0) {
			logw(
				"can not get the ve version ,both 0xf0 and 0xe4 is 0, try to "
				"read again\n");
			printk(
				"can not get the ve version ,both 0xf0 and 0xe4 is 0, try to "
				"read again\n");
			context->ve_version = 0;

			// we should read the ic version again when other process's codec is
			// reseting the ve
			read_count++;
			if (read_count < 4) {
				mdelay(1);
				goto READ_IC_VERSION;
			} else
				return -1;
		}

		dec_enc_ip = ((unsigned long long)value_dec_ip << 32) |
					 ((unsigned long long)value_enc_ip);
		logv(
			"the get enc_dec_ip_version:%llx, "
			"gVeEnvironmentInfo.address_macc:%p\n",
			dec_enc_ip, context->reg_base);

		context->ve_version = dec_enc_ip;
		printk("-------version value = 0\n");
	} else {
		context->ve_version = (unsigned long long)(value >> 16);
		printk("value>>16 ve_version = %lld\n", context->ve_version);
	}

	return 0;
}

static void* ve_aw_reg_group_addr(struct ve_interface* p, int group_type)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;
	char* ve_base = context->reg_base;
	char* ve_group_base;

	switch (group_type) {
	case REG_GROUP_VETOP:
		ve_group_base = ve_base + REG_GROUP_VETOP_OFFSET;
		break;
	case REG_GROUP_MPEG_DECODER:
		ve_group_base = ve_base + REG_GROUP_MPEG_OFFSET;
		break;
	case REG_GROUP_H264_DECODER:
		ve_group_base = ve_base + REG_GROUP_H264_OFFSET;
		break;
	case REG_GROUP_AVS_DECODER:
		ve_group_base = ve_base + REG_GROUP_AVS_OFFSET;
		break;
	case REG_GROUP_VC1_DECODER:
		ve_group_base = ve_base + REG_GROUP_VC1_OFFSET;
		break;
	case REG_GROUP_RV_DECODER:
		ve_group_base = ve_base + REG_GROUP_RV_OFFSET;
		break;
	case REG_GROUP_H265_DECODER:
		ve_group_base = ve_base + REG_GROUP_H265_OFFSET;
		break;
	case REG_GROUP_AVS2_DECODER:
		ve_group_base = ve_base + REG_GROUP_AVS2_OFFSET;
		break;
	case REG_GROUP_JPEG_DECODER:
		ve_group_base = ve_base + REG_GROUP_JPEG_OFFSET;
		break;
	default:
		ve_group_base = NULL;
		loge("group type not match!!");
	}
	//	logd("group addr:%p group type:%d base:%p",ve_group_base, group_type,
	//ve_base);
	return (void*)ve_group_base;
}

static unsigned long long ve_aw_get_ic_version(struct ve_interface* p)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;

	return context->ve_version;
}

static unsigned int ve_aw_get_phy_offset(struct ve_interface* p)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;

	return context->phy_offset;
}

static void ve_aw_set_ddr_mode(struct ve_interface* p, int type)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;

	vetop_reg_mode_sel_t* mode_reg =
		(vetop_reg_mode_sel_t*)(context->reg_base + TOP_MODE_SELECT_OFFSET);
	switch (type) {
	case DDRTYPE_DDR1_16BITS:
		mode_reg->ddr_mode = 0;
		break;
	case DDRTYPE_DDR1_32BITS:
	case DDRTYPE_DDR2_16BITS:
		mode_reg->ddr_mode = 1;
		break;
	case DDRTYPE_DDR2_32BITS:
	case DDRTYPE_DDR3_16BITS:
		mode_reg->ddr_mode = 2;
		break;
	case DDRTYPE_DDR3_32BITS:
	case DDRTYPE_DDR3_64BITS:
		mode_reg->ddr_mode = 3;
		mode_reg->rec_wr_mode = 1;
		break;
	default:
		logd("ddr type not match!");
		break;
	}

	logd("ji***set ddr mode:%x", *mode_reg);
}

static int ve_aw_set_freq(struct ve_interface* p, int freq)
{
	cedar_set_ve_freq(freq);
	return 0;
}

static void enable_encoder(struct ve_aw_ctx* context)

{
	vetop_reg_mode_sel_t* mode_reg =
		(vetop_reg_mode_sel_t*)(context->reg_base + TOP_MODE_SELECT_OFFSET);
	logd("ji**************enable ve");
	if (context->ve_version == 0x1623 || context->ve_version == 0x1625 ||
		context->ve_version == 0x1651 || context->ve_version == 0x1663)
		mode_reg->mode = 0xb;
	else {
		mode_reg->enc_isp_enable = 1;
		if (context->config.just_isp_enable == 0)
			mode_reg->enc_enable = 1;
	}
}

static void disable_encoder(struct ve_aw_ctx* context)
{
	vetop_reg_mode_sel_t* mode_reg =
		(vetop_reg_mode_sel_t*)(context->reg_base + TOP_MODE_SELECT_OFFSET);
	logd("ji**************disable ve");

	if (context->ve_version == 0x1623 || context->ve_version == 0x1625 ||
		context->ve_version == 0x1651 || context->ve_version == 0x1663)
		mode_reg->mode = 0x7;
	else {
		mode_reg->enc_isp_enable = 0;
		if (context->config.just_isp_enable == 0)
			mode_reg->enc_enable = 0;
	}
}

static void enable_decoder(struct ve_aw_ctx* context)
{
	volatile vetop_reg_mode_sel_t* mode_reg =
		(vetop_reg_mode_sel_t*)(context->reg_base + TOP_MODE_SELECT_OFFSET);

	// hardware no regs to set enable flags, we just set mode regs
	switch (context->config.codec_format) {
	case VIDEO_CODEC_H264:
		mode_reg->mode = 1;
		break;
	case VIDEO_CODEC_VP8:
		mode_reg->mode = 1;
		break;
	case VIDEO_CODEC_AVS:
		mode_reg->mode = 1;
		break;
	case VIDEO_CODEC_WMV3:
		mode_reg->mode = 2;
		break;
	case VIDEO_CODEC_RX:
		mode_reg->mode = 3;
		break;
	case VIDEO_CODEC_H265:
		mode_reg->mode = 4;
		break;
	case VIDEO_CODEC_MJPEG: {
		if (policy_list[context->ve_list_num].sole_jdec)
			mode_reg->jpg_dec_en = 1;
		else {
			mode_reg->mode = 0;
		}
		break;
	}
	case VIDEO_CODEC_AVS2:
		mode_reg->mode = 4;
		break;
	case VIDEO_CODEC_VP9:
		mode_reg->mode = 4;
		break;
	default:
		loge("codec format not support!");
		mode_reg->mode = 0xf;
		break;
	}
}

static void disable_decoder(struct ve_aw_ctx* context)
{
	volatile vetop_reg_mode_sel_t* mode_reg =
		(vetop_reg_mode_sel_t*)(context->reg_base + TOP_MODE_SELECT_OFFSET);
	logd("ji***disable decoder");
	mode_reg->mode = 7;
}

static void ve_aw_enable_ve(struct ve_interface* p)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;

	if (context->config.dec_flag == 1)
		enable_decoder(context);
	if (context->config.enc_flag == 1)
		enable_encoder(context);
}

static void ve_aw_disable_ve(struct ve_interface* p)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;

	if (context->config.dec_flag == 1)
		disable_decoder(context);
	if (context->config.enc_flag == 1)
		disable_encoder(context);
}

static void reset_encoder(struct ve_aw_ctx* context)
{
	vetop_reg_reset_t* reset_reg =
		(vetop_reg_reset_t*)(context->reg_base + TOP_VE_RESET_OFFSET);

	if (policy_list[context->ve_list_num].sole_venc ||
		policy_list[context->ve_list_num].sole_jdec) {
		reset_reg->encoder_reset = 1;
		reset_reg->encoder_reset = 0;
		logd("sole venc reset");
	} else {
		cedar_reset_ve();
	}
}

static void reset_decoder(struct ve_aw_ctx* context)
{
	cedar_reset_ve();
	*((unsigned int*)(context->reg_base + 0x30)) = 0x0;
	*((unsigned int*)(context->reg_base + 0x40)) = 0x0;
	*((unsigned int*)(context->reg_base + 0x80)) = 0x0;
	*((unsigned int*)(context->reg_base + 0x84)) = 0x0;
	*((unsigned int*)(context->reg_base + 0x88)) = 0x0;
}

static void ve_aw_reset_ve(struct ve_interface* p)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;
	//	loge("ji**********reset ve");
	if (context->config.dec_flag == 1)
		reset_decoder(context);
	if (context->config.enc_flag == 1)
		reset_encoder(context);
}

static int ve_aw_wait_interrupt(struct ve_interface* p)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;
	return cedar_wait_interrupt(context->config.dec_flag, 1);
}

static int get_chip_id_str(struct ve_aw_ctx* context, char* buf, int len)
{
	if (len < 10) {
		loge("buf len not enough");
		return -1;
	}
	strcpy(buf, "00003400");
	// todo: get chip id
	return 0;
}

struct ve_interface* ve_aw_create(struct ve_config config)
{
	struct ve_aw_ctx* context;
	int policy_size;
	int i;

	context = kmalloc(sizeof(struct ve_aw_ctx), GFP_KERNEL);
	if (context == NULL) {
		loge("create ion allocator failed, out of memory \n");
		return NULL;
	}

	memset(context, 0, sizeof(struct ve_aw_ctx));

	cedar_open();

	context->reg_base = cedar_get_ve_base_addr();
	if (context->reg_base == NULL) {
		loge("get reg base fail!");
		goto ERR;
	}

	context->interface.get_reg_group_addr = ve_aw_reg_group_addr;
	context->interface.set_ddr_mode = ve_aw_set_ddr_mode;
	context->interface.set_ve_freq = ve_aw_set_freq;
	context->interface.enable_ve = ve_aw_enable_ve;
	context->interface.disable_ve = ve_aw_disable_ve;
	context->interface.reset = ve_aw_reset_ve;
	context->interface.wait_interrupt = ve_aw_wait_interrupt;
	context->interface.get_ic_version = ve_aw_get_ic_version;
	context->interface.get_phy_offset = ve_aw_get_phy_offset;

	context->config = config;

	if ((context->config.dec_flag == 1 && context->config.enc_flag == 1) ||
		(context->config.dec_flag == 0 && context->config.enc_flag == 0)) {
		loge("this model not support both dec and enc");
		goto ERR;
	}

	if (get_ve_hardware_version(context) != 0) {
		loge("get ve version fail!");
		goto ERR;
	}

	if (context->ve_version == 0x1639) {
		context->phy_offset = 0x20000000;
	} else if (context->ve_version <= 0x1718) {
		context->phy_offset = 0x40000000;
	} else {
		context->phy_offset = 0;
	}

	if (get_chip_id_str(context, context->soc_str, 20) != 0) {
		loge("get chip id string fail");
	}

	// get plicy list index
	context->ve_list_num = -1;
	policy_size = sizeof(policy_list) / sizeof(struct ve_lock_policy);

	for (i = 0; i < policy_size; i++) {
		if (context->ve_version == policy_list[i].id &&
			(!strcmp(context->soc_str, policy_list[i].soc_data) ||
			 !strncmp(context->soc_str, policy_list[i].soc_data, 8))) {
			context->ve_list_num = i;
			break;
		}
	}
	logd("ji*****ve list index:%d", context->ve_list_num);
	if (context->ve_list_num == -1) {
		loge("get ve list index fail");
	}

	//* set default ve freq
	cedar_set_ve_freq(300);

	return &(context->interface);
ERR:
	kfree(context);
	return NULL;
}

void ve_aw_destory(struct ve_interface* p)
{
	struct ve_aw_ctx* context = (struct ve_aw_ctx*)p;

	if (context == NULL) {
		logd("you destroy mem but p is null");
		return;
	}

	cedar_close();

	kfree(context);
}
